--[[
Bison - a framework for free positioning of buffs.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:
	
Free Software Foundation, I.,
51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
--]]

local MAJOR = 'Bison'
local Addon = LibStub( 'AceAddon-3.0'):NewAddon( MAJOR, 'AceEvent-3.0', 'AceConsole-3.0', 'AceHook-3.0', 'LibDebugLog-1.0')
if not Addon then return end -- already loaded and no upgrade necessary

local L      = LibStub( 'AceLocale-3.0'):GetLocale( MAJOR)
local AceGUI = LibStub( 'AceGUI-3.0')
local LBF    = LibStub( 'LibButtonFacade', true)

Addon.name, Addon.localizedname = GetAddOnInfo( MAJOR)
Addon.version = format( '%s: %s |cFF20FF20(Cata)|r', L.Version, GetAddOnMetadata( MAJOR, 'Version') or 'unknown')

local DEFAULTS = { 
	profile = {
		['*']  = { xPos = 400, yPos = 300, xPadding = 5, yPadding = 15, scale = 1.0, rows = 1, cols = 16, number = 16, show = true, horizontal = true, sort = 'dec',   style = {}, },
		buff   = { xPos = 400, yPos = 400, xPadding = 5, yPadding = 15, scale = 1.0, rows = 2, cols = 20, number = 40, show = true, horizontal = true, sort = 'dec',   style = {}, },
		secure = { xPos = 400, yPos = 350, xPadding = 5, yPadding = 15, scale = 1.0, rows = 2, cols = 20, number = 40, show = true, horizontal = true, sort = 'dec',   style = {}, },
		hijack = { xPos = 400, yPos = 300, xPadding = 5, yPadding = 15, scale = 1.0, rows = 2, cols = 20, number = 40, show = true, horizontal = true, sort = 'dec',   style = {}, },
		debuff = { xPos = 400, yPos = 250, xPadding = 5, yPadding = 15, scale = 1.0, rows = 1, cols = 16, number = 16, show = true, horizontal = true, sort = 'dec',   style = {}, },
		weapon = { xPos = 400, yPos = 200, xPadding = 5, yPadding = 15, scale = 1.0, rows = 1, cols =  3, number =  3, show = true, horizontal = true, sort = 'none',  style = {}, },
		prog   = { xPos = 400, yPos = 150, xPadding = 5, yPadding = 15, scale = 1.0, rows = 1, cols = 16, number = 16, show = true, horizontal = true, sort = 'alpha', style = {}, },
		hidden = { xPos = 400, yPos = 100, xPadding = 5, yPadding = 15, scale = 1.0, rows = 2, cols =  8, number = 16, show = true, horizontal = true, sort = 'dec',   style = {}, },
		join     = false,
		blizzard = false,
		locked   = true,
		enabled  = true,
		timer    = true,
		flashing = true, 
	} 
}
local NOTHING = {}

local inCombat
local auras = {
	buff = {},
	debuff = {},
	weapon = {}
}
local options = {
	type = 'group', name = Addon:GetName(), childGroups = 'tree', args = {}
}

--[[---------------------------------------------------------------------------------
Local
------------------------------------------------------------------------------------]]
local function InitBuffs()
	for i = 1,40 do
		tinsert( auras.buff, { id = i })
	end
	for i = 1,16 do
		tinsert( auras.debuff, { id = i })
	end
	for i = 1,3 do
		tinsert( auras.weapon, { id = i })
	end
end

local function UpdateAura( id, aura, filter, now)
	name, rank, texture, count, debuffType, duration, expiration, unitCaster, isStealable, consolidate, spellID = UnitAura( PlayerFrame.unit, id, filter)
	local timeLeft = name and expiration and (expiration - now) or nil
	aura.id      = id
	aura.filter  = filter
	aura.name    = name
	aura.rank    = rank
	aura.texture = texture
	aura.count   = count
	aura.type    = debuffType
	aura.duration    = duration
	aura.expiration  = expiration
	aura.timeLeft    = timeLeft and (timeLeft > 0) and timeLeft or nil
	aura.duration    = duration
	aura.unitCaster  = unitCaster
	aura.isStealable = isStealable
	aura.consolidate = consolidate
	aura.spellID     = spellID
end

local function UpdateWaepon( id, enchant, timeLeft, charges)
	timeLeft    = enchant and timeLeft and (timeLeft / 1000)
	local aura = auras.weapon[id]
	aura.id       = id
	aura.name     = enchant and tostring( id) or nil
	aura.count    = charges 
	aura.texture  = enchant and GetInventoryItemTexture( PlayerFrame.unit, 15 + id) 
	aura.timeLeft = timeLeft and (timeLeft > 0) and timeLeft or nil
end

local function UpdateEnchant()
	local mainEnchant, mainTimeLeft, mainCharges, offEnchant, offTimeLeft, offCharges, thrEnchant, thrTimeLeft, thrCharges = GetWeaponEnchantInfo()
	UpdateWaepon( 1, mainEnchant, mainTimeLeft, mainCharges)
	UpdateWaepon( 2, offEnchant,  offTimeLeft,  offCharges)
	UpdateWaepon( 3, thrEnchant,  thrTimeLeft,  thrCharges)
end

local function UpdateAllAuras()
	Addon:Debug( 'UpdateAllAuras')
	local now = GetTime()
	for id,aura in pairs( auras.buff) do
		UpdateAura( id, aura, 'HELPFUL', now)
	end
	for id,aura in pairs( auras.debuff) do
		UpdateAura( id, aura, 'HARMFUL', now)
	end
	UpdateEnchant()
end

local function UpdateButtonMax()
	BUFF_MAX_DISPLAY = 40
end

--[[---------------------------------------------------------------------------------
Class
------------------------------------------------------------------------------------]]
function Addon:OnInitialize()
	self:ToggleDebugLog( false)
	InitBuffs()
	self.db = LibStub( 'AceDB-3.0'):New( 'BisonDB', DEFAULTS, 'Default')
	self:SetEnabledState( self.db.profile.enabled)
	LibStub( 'AceConfig-3.0'):RegisterOptionsTable( 'BisonDialog', options)
	self:RegisterChatCommand( 'bison', 'ShowDialog')
	self:RegisterChatCommand( 'bi', 'ShowDialog')
	if LBF then
		LBF:RegisterSkinCallback( 'Bison', self.OnSkin, self)
	end
end

function Addon:OnEnable()
	self:Debug( 'OnEnable')
	self:RegisterEvent( 'PLAYER_REGEN_ENABLED')
	self:RegisterEvent( 'PLAYER_REGEN_DISABLED')
--	securecall( UpdateButtonMax)
	UpdateAllAuras()
	self:SecureHook( 'BuffFrame_UpdateAllBuffAnchors', 'UpdateAllBuffAnchors')
	self:SecureHook( 'DebuffButton_UpdateAnchors', 'UpdateDebuffButton')
	self:SecureHookScript( TemporaryEnchantFrame, 'OnUpdate', 'OnUpdateEnchantFrame')
	self:SecureHook( 'AuraButton_UpdateDuration', 'UpdateDuration')
	self:Print( self.version, 'loaded')
end

function Addon:OnDisable()
	self:Debug( 'OnDisable')
	self:UnregisterEvent( 'PLAYER_REGEN_ENABLED')
	self:UnregisterEvent( 'PLAYER_REGEN_DISABLED')
	self:Unhook( 'BuffFrame_UpdateAllBuffAnchors')
	self:Unhook( 'DebuffButton_UpdateAnchors')
	self:Unhook( 'TemporaryEnchantFrame_OnUpdate')
	self:Unhook( 'AuraButton_UpdateDuration')
end

function Addon:PLAYER_REGEN_ENABLED( event, ...)
	inCombat = false
end

function Addon:PLAYER_REGEN_DISABLED( event, ...)
	inCombat = true
end

function Addon:InCombat()
	return inCombat
end

function Addon:OnSkin( skin, gloss, backdrop, group, button, colors)
	local profile = group and self.db.profile[group]
	if profile then
		if type( profile.style) ~= 'table' then
			profile.style = {}
		end
		local style = profile.style
		style.skin     = skin
		style.gloss    = gloss
		style.backdrop = backdrop
		style.colors   = colors
	end
end

function Addon:UpdateAllBuffAnchors()
--	BuffFrame_UpdateAllBuffAnchors()
	UpdateAllAuras()
	for name,mod in self:IterateModules() do
		if mod.UpdateBuffAnchors then
			mod:UpdateBuffAnchors()
		end
	end
end

function Addon:UpdateDebuffButton( buttonName, index)
--	DebuffButton_UpdateAnchors( buttonName, index)
	for name,mod in self:IterateModules() do
		if mod.UpdateDebuffAnchors then
			mod:UpdateDebuffAnchors(  buttonName, index)
		end
	end
end

function Addon:OnUpdateEnchantFrame( elapsed)
--	TemporaryEnchantFrame_OnUpdate()
	UpdateEnchant()
	for name,mod in self:IterateModules() do
		if mod.UpdateEnchantAnchors then
			mod:UpdateEnchantAnchors()
		end
	end
end

function Addon:UpdateDuration( button, timeLeft)
--	AuraButton_UpdateDuration( button, timeLeft)
	if SHOW_BUFF_DURATIONS == "1" and timeLeft then
		local child = _G[button:GetName()..'Duration']
		child:SetText( self:DurationString( timeLeft))
	end
end

--[[---------------------------------------------------------------------------------
Main
------------------------------------------------------------------------------------]]
function Addon:CloneAura( dst, name)
	for _,aura in pairs( auras[name] or NOTHING) do
		tinsert( dst, aura)
	end 
end

function Addon:DurationString( seconds)
	if not self.db.profile.timer then
		return format( SecondsToTimeAbbrev( seconds))
	end
	if not seconds then
		return ''
	end
	local negative, time
	if seconds < 0 then 
		negative = '-'
		time = -seconds
	else
		negative = ''
		time = seconds
	end
	if time < 60 then 
		return format( '|cffffffff%s%ds|r', negative, time) 
	end
	local min = floor( time / 60)
	if time < 600 then 
		return format( '|cffffffff%s%d:%02d|r', negative, min, time % 60) 
	end
	if time < 3600 then 
		return format( '|cffffffff%s%dm|r', negative, min) 
	end
	return format( '|cffffffff%s%d:%02dh|r', negative, floor( min / 60), min % 60)
end

function Addon:ShowDialog()
	for name,mod in self:IterateModules() do
		options.args[name] = mod.GetOptionTable and mod:GetOptionTable() or nil
	end
	LibStub( 'AceConfigDialog-3.0'):Open( 'BisonDialog')
end

--[[---------------------------------------------------------------------------------
Layout
------------------------------------------------------------------------------------]]
AceGUI:RegisterLayout( 'BuffBar', function( content, children)
	local profile = Addon.db.profile[content.obj.proName]
	local xPadding = -profile.xPadding or 0
	local yPadding = -profile.yPadding or 0
	local horizontal = profile.horizontal
	local scale = profile.scale or 1
	local cols = profile.cols or 10
	local rows = profile.rows or 4
	local last, first
	local width = 0
	local height = 0
	local i = 1
	for k,child in pairs( children) do
		local frame = child.frame
		frame:ClearAllPoints()
		if last and first then
			if horizontal then
				if xPadding >= 0 then
					frame:SetPoint( 'LEFT', last, 'RIGHT', xPadding, 0)
				else
					frame:SetPoint( 'RIGHT', last, 'LEFT', xPadding, 0)
				end
			else
				if yPadding >= 0 then
					frame:SetPoint( 'BOTTOM', last, 'TOP', 0, yPadding)
				else
					frame:SetPoint( 'TOP', last, 'BOTTOM', 0, yPadding)
				end
			end
		elseif first then
			if horizontal then
				if yPadding >= 0 then
					frame:SetPoint( 'BOTTOM', first, 'TOP', 0, yPadding)
				else
					frame:SetPoint( 'TOP', first, 'BOTTOM', 0, yPadding)
				end
			else
				if xPadding >= 0 then
					frame:SetPoint( 'LEFT', first, 'RIGHT', xPadding, 0)
				else
					frame:SetPoint( 'RIGHT', first, 'LEFT', xPadding, 0)
				end
			end
			first = frame
		else
			if xPadding >= 0 then
				if yPadding >= 0 then
					frame:SetPoint( 'BOTTOMLEFT', content)
				else
					frame:SetPoint( 'TOPLEFT', content)
				end
			else
				if yPadding >= 0 then
					frame:SetPoint( 'BOTTOMRIGHT', content)
				else
					frame:SetPoint( 'TOPRIGHT', content)
				end
			end
			first = frame
		end
		last = frame
		if child.SetScale then
			child:SetScale( scale)
		end
		if child.DoLayout then
			child:DoLayout()
		end
		width  = max( width, frame.width or frame:GetWidth() or 0)
		height = max( height, frame.height or frame:GetHeight() or 0)
		i = i + 1
		if horizontal then
			if i > cols then
				i = 1
				last = nil
			end
		else
			if i > rows then
				i = 1
				last = nil
			end
		end
	end
	content.obj:LayoutFinished( scale * cols * (width + abs(xPadding)), scale * rows * (height + abs(yPadding)))
end)
