--[[
Bison - a framework for free positioning of buffs.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:
	
Free Software Foundation, I.,
51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
--]]

local MAJOR = 'prototype'

local Addon  = LibStub( 'AceAddon-3.0'):GetAddon( 'Bison')
local L      = LibStub( 'AceLocale-3.0'):GetLocale( 'Bison')
local AceGUI = LibStub( 'AceGUI-3.0')
local LBF    = LibStub( 'LibButtonFacade', true)

local NOTHING = {}
local SORT_DESC = {
	[1] = L.SortNone,
	[2] = L.SortAlpha,
	[3] = L.SortRevert,
	[4] = L.SortInc,
	[5] = L.SortDec,
	[6] = L.SortDurationAsc,
	[7] = L.SortDurationDesc
}
local SORT_TYPES = {
	['none']   = 1,
	['alpha']  = 2,
	['revert'] = 3,
	['inc']    = 4,
	['dec']    = 5,
	['durationasc']  = 6,
	['durationdesc'] = 7,
}

--[[---------------------------------------------------------------------------------
Sorting
------------------------------------------------------------------------------------]]
local function SortNone( a, b)
	return a.id < b.id
end

local function SortNameAsc( a, b)
	if a.name and b.name then return a.name < b.name end
	if a.name then return true end
	if b.name then return false end
	return a.id < b.id
end

local function SortNameDesc( a, b)
	if a.name and b.name then return a.name > b.name end
	if a.name then return true end
	if b.name then return false end
	return a.id < b.id
end

local function SortTimeLeftAsc( a, b)
	if a.name and b.name then
		if a.timeLeft and b.timeLeft then return a.timeLeft < b.timeLeft end
		if a.timeLeft then return true end
		if b.timeLeft then return false end
		return a.name < b.name
	end
	if a.name then return true end
	if b.name then return false end
	return a.id < b.id
end

local function SortTimeLeftDesc( a, b)
	if a.name and b.name then
		if a.timeLeft and b.timeLeft then return a.timeLeft > b.timeLeft end
		if a.timeLeft then return false end
		if b.timeLeft then return true end
		return a.name < b.name
	end
	if a.name then return true end
	if b.name then return false end
	return a.id < b.id
end

local function SortDurationAsc( a, b)
	if a.name and b.name then
		if a.duration and b.duration then return a.duration < b.duration end
		if a.duration then return true end
		if b.duration then return false end
		return a.name < b.name
	end
	if a.name then return true end
	if b.name then return false end
	return a.id < b.id
end

local function SortDurationDesc( a, b)
	if a.name and b.name then
		if a.duration and b.duration then return a.duration > b.duration end
		if a.duration then return true end
		if b.duration then return false end
		return a.name < b.name
	end
	if a.name then return true end
	if b.name then return false end
	return a.id < b.id
end

local SORT = {
	['none']   = SortNone,
	['alpha']  = SortNameAsc,
	['revert'] = SortNameDesc,
	['inc']    = SortTimeLeftAsc,
	['dec']    = SortTimeLeftDesc,
	['durationasc']  = SortDurationAsc,
	['durationdesc'] = SortDurationDesc,
}

--[[---------------------------------------------------------------------------------
Script
------------------------------------------------------------------------------------]]
local function Bar_OnMouseDown( group, event, button)
	Addon:Debug( 'Bar_OnMouseDown', group.proName, button)
	group.frame:StartMoving()
end

local function Bar_OnMouseUp( group)
	Addon:Debug( 'Bar_OnMouseUp', group.proName)
	group.frame:StopMovingOrSizing()
	local profile = Addon.db.profile[group.proName]
	profile.xPos = group.frame:GetLeft()
	profile.yPos = group.frame:GetBottom()
end

local function Button_OnMouseDown( btn, event, button)
	btn.parent:Fire( 'OnMouseDown', button)
end

local function Button_OnMouseUp( btn)
	btn.parent:Fire( 'OnMouseUp')
end

local function CreateGroup( name, color, type)
	local group = AceGUI:Create( type)
	group:SetLayout( 'BuffBar')
	group.proName = name
	group:SetGhostColor( color)
	group:SetCallback( 'OnMouseDown', Bar_OnMouseDown)
	group:SetCallback( 'OnMouseUp',   Bar_OnMouseUp)
	group.frame:SetClampedToScreen( true )
	return group
end

--[[---------------------------------------------------------------------------------
Main
------------------------------------------------------------------------------------]]
local prototype = {
	['OnInitialize'] = function( self)
		self.profile = Addon.db.profile[self.proName]
		self.aura = {}
		self.group = CreateGroup( self.proName, self.color, self.grpType)
		if type( self.OnModuleInitialize) == 'function' then
			self:OnModuleInitialize()
		end
	end,
	
	['OnEnable'] = function( self)
		self:UpdateGroup()
		self:RegisterMessage( 'BISON_UPDATE')
		self:RegisterMessage( 'BISON_PROFILE')
		if type( self.OnModuleEnable) == 'function' then
			self:OnModuleEnable()
		end
		if LBF then
			local style = self.profile.style or NOTHING
			LBF:Group( 'Bison', self.proName):Skin( style.skin, style.gloss, style.backdrop, style.colors)
		end
		self.group:SetPos( self.profile.xPos, self.profile.yPos)
		self:SendMessage( 'BISON_UPDATE', self.proName)
	end,
	
	['OnDisable'] = function( self)
		self:SetVisible( false)
		self:UnregisterMessage( 'BISON_UPDATE')
		self:UnregisterMessage( 'BISON_PROFILE')
		if type( self.OnModuleDisable) == 'function' then
			self:OnModuleDisable()
		end
	end,

	['BISON_UPDATE'] = function( self, event, bar)
		if bar == nil or bar == self.proName then
			Addon:Debug( event, bar)
			self:UpdateGroup()
			self:SetVisible( not (self.hideOnLock and Addon.db.profile.locked))
		end
	end,
	
	['BISON_PROFILE'] = function( self)
		self.profile = Addon.db.profile[self.proName]
		if LBF then
			local style = self.profile.style or NOTHING
			LBF:Group( 'Bison', self.proName):Skin( style.skin, style.gloss, style.backdrop, style.colors)
		end
	end,

	['SetVisible'] = function( self, visible)
		Addon:Debug( 'SetVisible', visible, self.profile.show)
		if visible and self.profile.show then
			self.group:SetVisible( true)
			self:UpdateValues( true, false)
		else
			self.group:SetVisible( false)
			self:UpdateValues( false, false)
		end
	end,

	['CloneAura'] = function( self, name)
		Addon:CloneAura( self.aura, name)
	end,

	['RegisterOptions'] = function( self, options, menuName)
		local modName = 'Bison_'..self.proName
		LibStub( 'AceConfig-3.0'):RegisterOptionsTable( modName, options)
		LibStub( 'AceConfigDialog-3.0'):AddToBlizOptions( modName, menuName, self.proName)
	end,

	['Renumber'] = function( self)
		local group = self.group
		local max = self.profile.number
		while #group.children > max do
			local child = tremove( group.children)
			AceGUI:Release( child)
		end
		local n = #group.children + 1
		for i = n,max do
			local child = self:CreateButton()
			child:Initialize( i, self.filter)
			group:AddChild( child)
		end
		group:DoLayout()
	end,

	['UpdateGroup'] = function( self)
--		Addon:Debug( 'UpdateGroup', Addon.db.profile.locked)
		self:Renumber()
		local locked = Addon.db.profile.locked
		if locked then
			self.group:SetPos( self.profile.xPos, self.profile.yPos)
		end
		self.group:SetLock( locked or not self.profile.show)
	end,

	['UpdateChild'] = function( self, child, aura)
		local profile = self.profile
		child:SetUserData( 'id',   aura.id)
		child:SetUserData( 'name', aura.name)
		child:SetText( aura.timeLeft and (SHOW_BUFF_DURATIONS == '1') and Addon:DurationString( aura.timeLeft) or nil, aura.warn)
		child:SetSpell( profile.spell and aura.spellID)
		child:SetCount( aura.count)
		child:SetTexture( aura.texture)
		if profile.flashing and aura.timeLeft then
			if aura.duration and aura.duration < BUFF_WARNING_TIME then
				child:SetFlashing( aura.timeLeft < aura.duration/2)
			else
				child:SetFlashing( aura.timeLeft < BUFF_WARNING_TIME)
			end
		else
			child:SetFlashing( false)
		end
	end,

	['UpdateLBF'] = function( self, frame)
		if LBF and frame and not frame.LBF then
			LBF:Group( 'Bison', self.proName):AddButton( frame)
			frame.LBF = true
		end
	end,
	
	['CreateNormButton'] = function( self, type)
		local child = AceGUI:Create( type)
		child:SetCallback( 'OnMouseDown', Button_OnMouseDown)
		child:SetCallback( 'OnMouseUp',   Button_OnMouseUp)
		return child
	end,

	['InCombat'] = function( self)
		return Addon:InCombat()
	end,
	
	['IsLock'] = function( self)
		return Addon.db.profile.locked
	end,
	
	['GetSort'] = function( self)
		local sort = self.profile.sort
		return SORT[sort] or SortNone
	end,

	['GetProperty'] = function( self, info)
		local key = info[#info]
		return self.profile[key]
	end,
	
	['SetProperty'] = function( self, info, value)
		local key = info[#info]
		self.profile[key] = value
		self:SendMessage( 'BISON_UPDATE', self.proName)
	end,
	
	['SetScale'] = function( self, info, value)
		value = tonumber(value)
		if value and value > 0 and value <= 2 then
			self.profile.scale = value
			self:SendMessage( 'BISON_UPDATE', self.proName)
		end
	end,
	
	['SetCols'] = function( self, info, value)
		local maxButton = info.option.max or 16
		value = tonumber(value)
		if value and value > 0 and value <= maxButton then
			value = floor( value)
			self.profile.cols = value
			self.profile.rows = ceil( self.profile.number / value)
			self:SendMessage( 'BISON_UPDATE', self.proName)
		end
	end,
	
	['SetRows'] = function( self, info, value)
		local maxButton = info.option.max or 16
		value = tonumber(value)
		if value and value > 0 and value <= maxButton then
			value = floor( value)
			self.profile.cols = ceil( self.profile.number / value)
			self.profile.rows = value
			self:SendMessage( 'BISON_UPDATE', self.proName)
		end
	end,
	
	['SetNumber'] = function( self, info, value)
		local maxButton = info.option.max or 16
		value = tonumber(value)
		if value and value > 0 and value <= maxButton then
			value = floor( value)
			self.profile.number = floor( value)
			self.profile.cols   = ceil( self.profile.number / self.profile.rows)
			self.profile.rows   = ceil( self.profile.number / self.profile.cols)
			self:SendMessage( 'BISON_UPDATE', self.proName)
		end
	end,
	
	['SetXPadding'] = function( self, info, value)
		local min = info.option.mix or -20
		local max = info.option.max or 20
		value = tonumber(value)
		if value and value >= min and value <= max then
			self.profile.xPadding = floor( value)
			self:SendMessage( 'BISON_UPDATE', self.proName)
		end
	end,
	
	['SetYPadding'] = function( self, info, value)
		local min = info.option.mix or -50
		local max = info.option.max or 50
		value = tonumber(value)
		if value and value >= min and value <= max then
			self.profile.yPadding = floor( value)
			self:SendMessage( 'BISON_UPDATE', self.proName)
		end
	end,
	
	['GetJoin'] = function( self)
		return Addon.db.profile.join
	end,
	
	['SetJoin'] = function( self, info, value)
		Addon.db.profile.join = value
		self:SendMessage( 'BISON_UPDATE', self.proName)
	end,
	
	['GetSortDesc'] = function( self)
		return SORT_DESC
	end,
	
	['GetSortType'] = function( self)
		return SORT_TYPES[self.profile.sort]
	end,
	
	['SetSortType'] = function( self, info, value)
		for k,v in pairs(SORT_TYPES) do
			if v == value then
				self.profile.sort = k
				self:SendMessage( 'BISON_UPDATE', self.proName)
				break
			end
		end
	end,
}
--[[---------------------------------------------------------------------------------
Class
------------------------------------------------------------------------------------]]
function Addon:NewBarModule( name, ...)
	if self:GetModule( name, true) then
		error( 'can not register double module: '..name, 2)
	end
	return self:NewModule( name, prototype, 'AceEvent-3.0', ...)
end

